# SPDX-FileCopyrightText: 2023-present Fynn Freyer <fynn.freyer@googlemail.com>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
Wrapper around the OpenAI API.
"""

# ruff: noqa: A003, RUF012

from __future__ import annotations

from itertools import chain
from json import dumps
from logging import getLogger
from typing import Optional

from openai import ChatCompletion
from sqlalchemy import Enum as SQLEnum
from sqlalchemy import ForeignKey
from sqlalchemy.orm import Mapped, mapped_column, relationship
from sqlalchemy.util import hybridproperty

from ai_chatter.db.base import Base
from ai_chatter.enums import ChatGPTModel, SenderRole
from ai_chatter.utils.config import SessionSettings

_logger = getLogger(__name__)


class Message(Base):
    """Represents a message to or from ChatGPT."""

    __tablename__ = "message"

    id: Mapped[int] = mapped_column(primary_key=True)
    """The message id."""

    role: Mapped[SenderRole] = mapped_column(SQLEnum(SenderRole))
    """The role of the sender. Can be one of "user", "assistant" or "system"."""

    content: Mapped[str]
    """The message content."""

    tokens: Mapped[Optional[int]]
    """The amount of tokens this message used up."""

    def to_dict(self):
        return {
            "role": self.role,
            "content": self.content,
        }

    __mapper_args__ = {
        "polymorphic_on": role,
        "polymorphic_abstract": True,
    }


class BehaviorDirective(Message):
    """Represents a message to ChatGPT containing a directive specifying wanted behavior."""

    __tablename__ = "behavior_directive"

    id: Mapped[int] = mapped_column(ForeignKey("message.id"), primary_key=True)
    """The :class:`Message` id."""

    # TODO make "unique" BehaviorDirectives via "upserts"

    __mapper_args__ = {
        "polymorphic_identity": SenderRole.SYSTEM,
    }


class Prompt(Message):
    """Represents a message to ChatGPT containing a prompt to complete."""

    __tablename__ = "prompt"

    id: Mapped[int] = mapped_column(ForeignKey("message.id"), primary_key=True)
    """The :class:`Message` id."""

    session_id: Mapped[int] = mapped_column(ForeignKey("session.id"))
    """The ID of the :class:`Session`, this prompt belongs to."""

    session: Mapped[Session] = relationship(back_populates="prompts")
    """The :class:`Session` this message belongs to."""

    response: Mapped[Optional[Response]] = relationship(foreign_keys="response.c.prompt_id")
    """The :class:`Response` that belongs to this prompt."""

    __mapper_args__ = {
        "polymorphic_identity": SenderRole.USER,
    }


class Response(Message):
    """Represents a message from ChatGPT, completing a prompt."""

    __tablename__ = "response"

    id: Mapped[int] = mapped_column(ForeignKey("message.id"), primary_key=True)
    """The :class:`Message` id."""

    prompt_id: Mapped[int] = mapped_column(ForeignKey("prompt.id"))
    """Id of the :class:`Prompt` that was completed by this response."""

    prompt: Mapped[Prompt] = relationship(foreign_keys="response.c.prompt_id", back_populates="response")
    """The :class:`Prompt` that was completed by this response."""

    raw_json: Mapped[str]
    """The raw JSON response."""

    @hybridproperty
    def session_id(self):
        """The ID of the :class:`Session`, this response belongs to."""
        return self.prompt.session_id

    @hybridproperty
    def session(self):
        """The the :class:`Session`, this response belongs to."""
        return self.prompt.session

    __mapper_args__ = {
        "polymorphic_identity": SenderRole.ASSISTANT,
    }


class Session(Base):
    """Represents a set of interactions with ChatGPT, e.g., some prompts and answers."""

    __tablename__ = "session"

    id: Mapped[int] = mapped_column(primary_key=True)
    """The session id."""

    model: Mapped[ChatGPTModel] = mapped_column(SQLEnum(ChatGPTModel), default=ChatGPTModel.GPT4)
    """The model to use. Defaults to GPT-4."""

    behavior_id: Mapped[Optional[int]] = mapped_column(ForeignKey("behavior_directive.id"))
    """
    A directive to ChatGPT indicating desired behavior, e.g., "Only output Python code, compatible with version3.8".
    Will be included in requests regardless of :attr:`context` settings.
    """

    behavior: Mapped[Optional[BehaviorDirective]] = relationship()
    """
    A directive to ChatGPT indicating desired behavior, e.g., "Only output Python code, compatible with version3.8".
    Will be included in requests regardless of :attr:`context` settings.
    """

    context_size: Mapped[int] = mapped_column(default=-1)
    """
    How many message pairs of the history to include as context. Negative numbers send complete history. Defaults to -1.
    """

    prompts: Mapped[list[Prompt]] = relationship()
    """The history of :class:`Prompt` objects sent to ChatGPT."""

    @hybridproperty
    def responses(self) -> list[Response]:
        """The history of :class:`Response` objects sent by ChatGPT."""
        return [prompt.response for prompt in self.prompts if prompt.response is not None]

    # TODO auto summarize feature, where we use ChatGPT to summarize the conversation so far, to compress context.

    def __init__(
        self, model: ChatGPTModel = ChatGPTModel.GPT4, context: int = 0, behavior: Optional[str] = None, **kwargs
    ):
        """Get a prompt from the user, send it to ChatGPT, and display the response."""
        super().__init__(**kwargs)

        self.model = model
        self.context_size = context
        if behavior is not None:
            self.behavior = BehaviorDirective(content=behavior)

    def complete(self, prompt: str) -> str:
        """
        Get a prompt completion.

        :param prompt: The prompt to send to ChatGPT.
        :return: The answer.
        """
        message = Prompt(
            content=prompt,
            session=self,
        )

        completion = ChatCompletion.create(
            model=self.model,
            messages=[*self.context, message.to_dict()],
        )

        message.tokens = completion["usage"]["prompt_tokens"]

        response = Response(
            role="assistant",
            content=completion["choices"][0]["message"]["content"],
            raw_json=dumps(completion),
            tokens=completion["usage"]["completion_tokens"],
            prompt=message,
        )

        return response.content

    @hybridproperty
    def context(self) -> list[dict]:
        """The context to send to ChatGPT."""
        if 0 < self.context_size <= len(self.answered):
            history_ctx = self.answered[: -self.context_size]
        elif self.context_size != 0:
            history_ctx = self.answered[:]
        else:
            history_ctx = []

        context = [] if self.behavior is None else [self.behavior.to_dict()]
        context.extend(msg.to_dict() for msg in chain.from_iterable(history_ctx))

        return context

    @hybridproperty
    def messages(self) -> list[Message]:
        """A flat list of messages."""
        return [msg for msg in chain.from_iterable(self.history) if msg is not None]

    @hybridproperty
    def history(self) -> list[tuple[Prompt, Optional[Response]]]:
        """The message history. It consists of :class:`Prompt` and :class:`Response` message pairs."""
        return [(prompt, prompt.response) for prompt in self.prompts]

    @hybridproperty
    def answered(self) -> list[tuple[Prompt, Response]]:
        """The :class:`Prompt` and :class:`Response` message pairs, that already have a response."""
        return [(prompt, response) for prompt, response in self.history if response is not None]

    @hybridproperty
    def tokens(self) -> int:
        """Number of tokens used up by the messages in this :class:`Session`."""
        return sum(msg.tokens for msg in self.messages if msg.tokens is not None)

    @classmethod
    def from_settings(cls, settings: SessionSettings):
        """Instantiate a session from :class:`ai_chatter.utils.config.SessionSettings`."""
        return cls(settings.model, settings.context_size_messages, settings.behavior)



# SPDX-FileCopyrightText: 2023-present Fynn Freyer <fynn.freyer@googlemail.com>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
This module defines the base class for ai_chatter applications.
"""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from logging import DEBUG, basicConfig, getLogger
from typing import Optional

import openai  # type: ignore
from sqlalchemy import Engine, create_engine
from sqlalchemy.orm import Session as DBSession

from ai_chatter.__about__ import __app_name__
from ai_chatter.communication import CLICommunicator, Communicator
from ai_chatter.db import Base, Session
from ai_chatter.utils.config import Settings

_logger = getLogger(__name__)

__all__ = ["Application"]


@dataclass
class Application(ABC):
    """
    The base class for ai_chatter applications. Subclass this, override the :meth:`run` method and then start your
    application with the :meth:`start` method.
    """

    settings: Settings
    session: Session
    communicator: Communicator
    engine: Optional[Engine] = None

    def __init__(
        self,
        settings: Optional[Settings] = None,
        session: Optional[Session] = None,
        communicator: Optional[Communicator] = None,
        engine: Optional[Engine] = None,
    ):
        self.settings = settings or Settings.load()

        if self.settings.api_key is None:
            msg = "No api key provided in settings."
            _logger.error(msg)
            raise ValueError(msg)

        self.session = session or Session(
            self.settings.session_settings.model, self.settings.session_settings.context_size_messages
        )

        self.communicator = communicator or CLICommunicator()

        if self.settings.persist:
            _logger.info(f"Persisting data for {__app_name__} in {self.settings.data_dir}.")

            self.settings.data_dir.mkdir(parents=True, exist_ok=True)
            db = self.settings.data_dir / "session.db"

            echo = self.settings.verbosity == DEBUG
            self.engine = engine or create_engine(f"sqlite:///{db}", echo=echo)
            Base.metadata.create_all(self.engine)
        else:
            self.engine = None

    def start(self):
        """Start the application."""
        basicConfig(level=self.settings.verbosity)
        _logger.info(f"Starting {__app_name__}.{self.__class__.__name__}")
        openai.api_key = self.settings.api_key
        if self.engine is not None:
            with DBSession(self.engine) as db_session:
                self.run()
                db_session.add(self.session)
                db_session.commit()
        else:
            self.run()

    @abstractmethod
    def run(self):
        """The method providing the main functionality of the application."""



# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

from ai_chatter.__about__ import __app_name__, __author__, __version__

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = __app_name__
copyright = f"2023-present, {__author__}"  # noqa: A001
author = __author__
release = __version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx_rtd_theme",
    "sphinxcontrib.cairosvgconverter",
    "sphinxcontrib.plantuml",
    "sphinxcontrib_relativeinclude",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- Options for output ------------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-latex-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["assets"]
templates_path = ["_templates"]

latex_elements = {
    "extraclassoptions": "openany,oneside",
    "papersize": "a4paper",
    "pointsize": "12pt",
    "figure_align": "H",
}

# -- Autodoc configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html
# https://www.sphinx-doc.org/en/master/usage/extensions/autosummary.html

add_module_names = False

autodoc_mock_imports = ["alembic", "sqlalchemy", "pydantic_settings", "docutils", "sphinx"]
autodoc_member_order = "bysource"

autosummary_context = {
    "project": __app_name__,
    "version": __version__,
}

# def setup(app):
#     """Register the directive."""
#     app.add_directive("relativeinclude", RelativeInclude)



# SPDX-FileCopyrightText: 2023-present Fynn Freyer <fynn.freyer@googlemail.com>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
This module defines the base class for ai_chatter applications.
"""

try:
    from bs4 import BeautifulSoup
    from markdownify import MarkdownConverter, markdownify
    from requests_html import HTMLSession
except ImportError as e:
    from ai_chatter.__about__ import __app_name__

    msg = f"Please install {__app_name__} with the optional web dependencies to enable this module. You can do so, by running `pip install {__app_name__}[web]` from the command line."
    raise ImportError(msg) from e

from ai_chatter.apps.base import Application


def md(soup: BeautifulSoup, **kwargs):
    """Shorthand for HTML -> Markdown conversion."""
    return MarkdownConverter(**kwargs).convert_soup(soup)


class SummarizePage(Application):
    """Summarize the contents of a webpage."""

    def __init__(self, url: str):
        super().__init__()
        self.url = url

    def run(self):
        """Scrape a web page and let ChatGPT provide a summary."""
        # get page and render js
        with HTMLSession() as session:
            response = session.get(self.url, allow_redirects=True)
            response.html.render()
            html = response.html.raw_html

        # cleanup content
        soup = BeautifulSoup(html, "html.parser")
        main = soup.find("main")
        if main is not None:
            soup = main
        remove_tags = ["script", "style", "header", "footer", "nav", "input", "textarea", "img"]
        for remove_tag in remove_tags:
            for tag in soup.select(remove_tag):
                tag.decompose()
        for code in soup.select("code"):
            code.unwrap()

        # turn into markdown
        raw_markdown = markdownify(soup.text).strip()

        # filter consecutive empty lines
        lines = []
        was_empty = True
        for line in raw_markdown.splitlines():
            is_empty = line.strip() == ""
            if not (is_empty and was_empty):
                lines.append(line)
            was_empty = is_empty

        markdown = "\n".join(lines)
        print(markdown)  # noqa



# SPDX-FileCopyrightText: 2023-present Fynn Freyer <fynn.freyer@googlemail.com>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
This module defines entry points for CLI scripts.
"""

import sys

from ai_chatter.apps.chat import Chatter
from ai_chatter.apps.howto import ShellHowTo
from ai_chatter.apps.web import SummarizePage
from ai_chatter.utils.config import Settings


def main():
    """The main entrypoint into the application."""
    settings = Settings.from_args()
    app = Chatter(settings)
    app.start()


def howto():
    """Entrypoint for the howto script."""
    app = ShellHowTo(sys.argv[1:])
    app.start()


def summarize():
    # parser = Settings.get_parser("summarize", "Summarize a webpage with ChatGPT.")
    # parser.add_argument("url", help="The URL of the webpage to summarize.")
    # args = Settings.parse_args(parser)
    # url = args.url

    app = SummarizePage(
        "https://stackoverflow.com/questions/3711856/how-to-remove-empty-lines-with-or-without-whitespace-in-python"
    )
    app.start()


if __name__ == "__main__":
    summarize()
    # main()



# SPDX-FileCopyrightText: 2023-present Fynn Freyer <fynn.freyer@googlemail.com>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
Module for classes and functions related to configuration.
"""

from __future__ import annotations

import sys
from argparse import ArgumentParser, Namespace
from json import JSONDecodeError, load
from logging import DEBUG, ERROR, INFO, WARNING, getLogger
from pathlib import Path
from typing import Literal, Optional

import platformdirs
from platformdirs import PlatformDirs
from pydantic import DirectoryPath, Field, NewPath
from pydantic_settings import BaseSettings, SettingsConfigDict

from ai_chatter.__about__ import __app_name__, __author__, __description__, __version__

__all__ = ["Settings", "SessionSettings"]

from ai_chatter.enums import ChatGPTModel

_logger = getLogger(__name__)


class SessionSettings(BaseSettings):
    """Settings of a :class:`ai_chatter.chat.Session`."""

    model: ChatGPTModel = ChatGPTModel.GPT4
    """The model to use. Defaults to GPT-4."""

    context_size_tokens: int = 8000
    """
    Maximum amount of (estimated) tokens of the message history to include as context. Negative numbers send complete
    history. Defaults to 8000.
    """

    context_size_messages: int = -1
    """
    How many message pairs of the history to include as context. Negative numbers send complete history. Defaults to -1.
    """

    session_warn_size: int = -1
    """
    After how many tokens should a warning message be printed out to the user. Negative numbers don't send warnings.
    Defaults to -1.
    """

    behavior: Optional[str] = None
    """A behavior directive to pass to ChatGPT."""


class Settings(BaseSettings):
    """Settings for the application. Can be set via env-vars or a config file."""

    model_config = SettingsConfigDict(env_prefix=f"{__app_name__}_")

    api_key: str = ""
    """The API key to use."""

    data_dir: DirectoryPath | NewPath = Path(platformdirs.user_data_dir(__app_name__, __author__))
    """The data directory of the application. E.g., "~/.local/share/ai_chatter"."""

    session_settings: SessionSettings = Field(default_factory=SessionSettings)
    """Settings concerning the session."""

    verbosity: Literal[ERROR, WARNING, INFO, DEBUG] = ERROR
    """How much information to print to screen."""

    persist: bool = True
    """Whether the conversation should be saved."""

    @classmethod
    def load(cls, file: Optional[str | Path] = None):
        """
        Load settings from the standard location or env vars. Location depends on the platform, e.g.,
        "${XDG_CONFIG_HOME}/ai_chatter/config.json" on Linux.

        :param file: Optionally, a configuration file to read from. Defaults to "config.json" in the current directory.
        :raise JSONDecodeError: If a config file contains invalid JSON.
        :raise RuntimeError: If no API key was provided.
        :return: A settings instance.
        """

        file = Path(file).resolve() if file is not None else Path("config.json").resolve()

        dirs = PlatformDirs(appname=__app_name__, appauthor=__author__)
        config_paths = [dirs.site_config_path / "config.json", dirs.user_config_path / "config.json", file]

        data_dir = dirs.user_data_path
        config = Settings(data_dir=data_dir).model_dump()

        for config_path in config_paths:
            _logger.info(f"{cls.__name__}: Trying to load config from {config_path}.")
            if config_path.exists():
                _logger.info(f"{cls.__name__}: Found file at {config_path}.")
                try:
                    with config_path.open() as config_file:
                        config |= load(config_file)
                except JSONDecodeError as e:
                    msg = f"{cls.__name__}: Invalid JSON in config file at {config_path}."
                    _logger.error(msg)
                    raise JSONDecodeError(msg, e.doc, e.pos) from e

        return cls(**config)

    @staticmethod
    def get_parser(name: str = __app_name__, description: str = __description__) -> ArgumentParser:
        """Get an ArgumentParser."""
        parser = ArgumentParser(name, description=description)

        parser.add_argument("-a", "--api-key", default=None, help="The API key to use.")
        parser.add_argument("-c", "--config", type=Path, default=None, help="Path to a config file.")
        parser.add_argument(
            "-e",
            "--ephemeral",
            dest="persist",
            action="store_false",
            default=True,
            help="Don't persist the conversation.",
        )
        parser.add_argument("-l", "--log", type=Path, default=None, help="Path to a log file.")

        parser.add_argument(
            "-v", "--verbose", dest="verbosity", action="count", default=0, help="Increase the level of verbosity."
        )
        parser.add_argument(
            "-V", "--version", action="version", version=__version__, help="Print version information and exit"
        )

        return parser

    @classmethod
    def parse_args(cls, parser: Optional[ArgumentParser] = None, args: Optional[tuple[str, ...]] = None) -> Namespace:
        """Parse arguments."""
        if args is None:
            args = tuple(sys.argv[1:])
        if parser is None:
            parser = cls.get_parser()

        parsed_args = parser.parse_args(args)
        verbosity = [ERROR, WARNING, INFO, DEBUG][min(parsed_args.verbosity, 3)]
        parsed_args.verbosity = verbosity

        return parsed_args

    @classmethod
    def from_args(cls, args: Optional[Namespace] = None):
        """Augment the settings object provided by :meth:`load` with command line parameters."""
        if args is None:
            args = cls.parse_args()
        settings = cls.load(args.config)

        settings.verbosity = args.verbosity
        if args.api_key is not None:
            settings.api_key = args.api_key

        return settings



# SPDX-FileCopyrightText: 2023-present Fynn Freyer <fynn.freyer@googlemail.com>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""

"""

from __future__ import annotations

import platform
from dataclasses import dataclass

from ai_chatter.apps.base import Application
from ai_chatter.db import BehaviorDirective
from ai_chatter.utils.config import Settings

__all__ = ["ShellHowTo"]


@dataclass
class EnvData:
    shell: str
    platform: str


class ShellHowTo(Application):
    """An :class:`ai_chatter.apps.base.Application` to get a commando for the current shell."""

    def __init__(self, args: list[str]):
        super().__init__()
        system_info = self.find_env_data()
        self.session.behavior = BehaviorDirective(
            content=f"Only return a shell command, or a series of commands delineated by newlines. Don't give "
                    f"explanations or further information. The commands should be runnable directly.\n"
                    f"I'm running a {system_info} system."
        )
        self.prompt = " ".join(args)

    def find_env_data(self) -> str:
        """Find data on the current platform to provide context for ChatGPT."""
        # TODO find shell information
        system = platform.system()
        if system == "Linux" or "BSD" in system:
            try:
                with open("/etc/os-release") as release:
                    prefix = "PRETTY_NAME="
                    for line in release:
                        if line.upper().startswith(prefix):
                            return line[len(prefix):].strip().strip('"')
            except FileNotFoundError:
                # /etc/os-release is not available
                return system

        return f"{system} {platform.version()}"

    def run(self):
        """Get a shell command from a question."""
        response = self.session.complete(self.prompt)
        self.communicator.show_response(response)


class LangHowTo(Application):
    """An :class:`ai_chatter.apps.base.Application` to solve a problem in a particular programming language."""

    def __init__(self, settings: Settings, lang: str):
        super().__init__(settings)
        self.lang = lang

    # def run(self):
    #     self.communicator



# SPDX-FileCopyrightText: 2023-present Fynn Freyer <fynn.freyer@googlemail.com>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

__app_name__ = "ai-chatter"
__description__ = "Toolkit for working with ChatGPT."
__version__ = "0.0.1a0"
__author__ = "Fynn Freyer"



# SPDX-FileCopyrightText: 2023-present Fynn Freyer <fynn.freyer@googlemail.com>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""

"""

from abc import ABC, abstractmethod
from typing import Optional

__all__ = ["Communicator"]


class Communicator(ABC):
    """Base class for getting prompts from a user, and displaying the responses."""

    @abstractmethod
    def get_prompt(self) -> Optional[str]:
        """Generate a prompt. Returning ``None`` indicates that the Session should end."""

    @abstractmethod
    def show_response(self, response: str):
        """Display the response."""



# SPDX-FileCopyrightText: 2023-present Fynn Freyer <fynn.freyer@googlemail.com>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
This submodule holds database-related classes and functions that are of cross-cutting concern.
"""

from datetime import datetime

from sqlalchemy import Engine, event
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, validates
from sqlalchemy.sql import func

__all__ = ["Base"]


class Base(DeclarativeBase):  # pylint: disable=not-callable
    """The declarative base for the ORM. Adds ``created_at`` and ``updated_at`` timestamp columns."""

    created_at: Mapped[datetime] = mapped_column(server_default=func.now())
    """A :class:`datetime` object, that is set on creation."""

    updated_at: Mapped[datetime] = mapped_column(server_default=func.now(), server_onupdate=func.now())  # type: ignore
    """A :class:`datetime` object, that is updated on changes."""

    @validates("created_at")
    def _validate_created_at(self, key, created_at: datetime):
        """
        Make sure that ``created_at`` timestamp isn't modified.

        :param key: Name of the modified column.
        :param settings: A :class:`rki_hivclustering.utils.config.HIVTraceSettings` object or a JSON dump representing one.
        :raises ValueError:  If settings is not a valid :class:`rki_hivclustering.utils.config.HIVTraceSettings` object, or JSON representation thereof.
        :return: A JSON encoded :class:`rki_hivclustering.utils.config.HIVTraceSettings`
        """
        # TODO turn this into an immutable decorator
        # if the created_at field already exists, we want to throw
        if self.created_at is not None:
            msg = f'{self.__class__.__name__}: column "{key}" cannot be modified retroactively.'
            raise ValueError(msg)
        return created_at


@event.listens_for(Engine, "connect")
def sqlite_pragma_enforce_foreign_keys(dbapi_connection, connection_record):  # noqa: ARG001
    """
    This is necessary to enforce foreign key constraints on SQLite.
    Cf. `SQLAlchemy docs <https://docs.sqlalchemy.org/en/20/dialects/sqlite.html#foreign-key-support>`
    """
    cursor = dbapi_connection.cursor()
    cursor.execute("PRAGMA foreign_keys=ON")
    cursor.close()



# SPDX-FileCopyrightText: 2023-present Fynn Freyer <fynn.freyer@googlemail.com>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""

"""

from typing import Optional

from ai_chatter.communication.base import Communicator

__all__ = ["CLICommunicator"]


class CLICommunicator(Communicator):
    """Prompts the user for an input and prints responses to the screen."""

    def get_prompt(self) -> Optional[str]:
        prompt = input("> ")
        return prompt if prompt not in ["q", "quit"] else None

    def show_response(self, response: str) -> None:
        print(response, "\n")  # noqa: T201



# SPDX-FileCopyrightText: 2023-present Fynn Freyer <fynn.freyer@googlemail.com>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""

"""

from ai_chatter.db.base import *
from ai_chatter.db.chat import *



# SPDX-FileCopyrightText: 2023-present Fynn Freyer <fynn.freyer@googlemail.com>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""
This module defines different enumerations for use as a controlled vocabulary.
"""

from enum import Enum
from functools import total_ordering

try:
    from enum import StrEnum  # type: ignore
except ImportError:

    class StrEnum(str, Enum):  # type: ignore
        """
        These can be used as drop-in for string based values, and can use normal formatting, etc.
        ``StrEnum`` is not part of the standard library in Python 3.8, so we roll our own.
        """


class SenderRole(StrEnum):
    """Attached to a :class:`ai_chatter.model.chat.Message`."""

    USER = "user"
    ASSISTANT = "assistant"
    SYSTEM = "system"


@total_ordering
class ChatGPTModel(StrEnum):
    """Specifies available standard models to use with ChatGPT."""

    GPT4 = "gpt-4"
    """The most current model."""

    GPT35_TURBO = "gpt-3.5-turbo"
    """The slightly older and cheaper model."""

    def __lt__(self, other):
        """Implement ordering via order of definitions."""
        if not isinstance(other, self.__class__):
            return NotImplemented
        defined = list(self.__class__.__members__.values())
        return defined.index(self) > defined.index(other)

    def __repr__(self):
        return str(self)



# SPDX-FileCopyrightText: 2023-present Fynn Freyer <fynn.freyer@googlemail.com>
#
# SPDX-License-Identifier: AGPL-3.0-or-later

"""

"""

from ai_chatter.apps.base import Application

__all__ = ["Chatter"]


class Chatter(Application):
    """A simple AI chatbot on your CLI."""

    def run(self):
        """Simple main loop, that takes a"""
        while (prompt := self.communicator.get_prompt()) is not None:
            response = self.session.complete(prompt)
            self.communicator.show_response(response)
